
#ifndef INC_4NEURO_DATASET_H
#define INC_4NEURO_DATASET_H

#include <iostream>
#include <fstream>
#include <utility>
#include <vector>
#include <map>
#include <string>
#include <functional>
#include <limits>
#include <memory>

#include "../settings.h"
#include "../NormalizationStrategy/NormalizationStrategy.h"

/* Forward declaration or arma::Mat<> type */
namespace arma {
    template<class T>
    class Mat;
}

namespace lib4neuro {
    /**
     * Class representing data, which can be used for training
     * and testing purposes.
     */
    class DataSet {

    protected:

        /**
         * Number of elements in the data set
         */
        size_t n_elements = 0;

        /**
         * Dimension of the input
         */
        size_t input_dim = 0;

        /**
         * Dimension of the output
         */
        size_t output_dim = 0;


        bool normalized = false;

        /**
         * Maximum (index 0) and minimum (index 1) input value
         */
        std::vector<double> max_min_inp_val;  //TODO make more efficiently, than by vector!

        /**
         * Stored data in the format of pairs of corresponding
         * input and output vectors
         */
        std::vector<std::pair<std::vector<double>, std::vector<double>>> data;

        std::vector<unsigned int> permutation_data_inds;

        /**
         *
         * @tparam T
         * @param v
         * @return
         */
        template<class T>
        std::vector<std::vector<T>> cartesian_product(const std::vector<std::vector<T>>* v);

        double get_vector_distance(
            const std::vector<double> &vec_a,
            const std::vector<double> &vec_b
        ) const;

        std::pair<std::vector<double>, std::vector<double>> aggregate_entries(
			const std::vector<std::vector<double>> & inputs,
			const std::vector<size_t> & sort_indices,
			const std::vector<size_t>::iterator lower_el_idx,
			const std::vector<size_t>::iterator upper_el_idx
        ) const;

        /**
         *
         */
        //TODO let user choose in the constructor!
        std::shared_ptr<NormalizationStrategy> normalization_strategy;

        arma::Mat<double>* inputs_matrix;

        arma::Mat<double>* outputs_matrix;

        arma::Mat<double>* stability_matrix;

        unsigned int current_batch_vector_ind = 0;

        /**
         * Indices of the active elements in DataSet
         *
         * If empty, than the whole DataSet is active!
         */
        std::vector<unsigned int> active_elements;

        unsigned int num_active_elements;


    public:

        /**
         * Struct used to access private properties from
         * the serialization function
         */
        struct access;

        /**
         * Constructor for an empty DataSet
         */
        LIB4NEURO_API DataSet();

        /**
         * Constructor reading data from the file
         * @param file_path Path to the file with stored data set
         */
        LIB4NEURO_API explicit DataSet(std::string file_path);

        /**
         * Constructor accepting data vector
         * @param data_ptr Pointer to the vector containing data
         */
        LIB4NEURO_API DataSet(const std::vector<std::pair<std::vector<double>, std::vector<double>>>* data_ptr,
                              NormalizationStrategy* ns = nullptr);

        /**
         * Constructs and empty data set of the given dimension
         */
        LIB4NEURO_API DataSet(size_t inp_dim, size_t out_dim,
                              NormalizationStrategy* ns = nullptr);

        /**
         * Creates a new data set with input values equidistantly positioned
         * over the certain interval and the output value
         * being constant
         *
         * Both input and output are 1-dimensional
         *
         * @todo add bounds as vectors for multi-dimensional data-sets
         *
         * @param lower_bound Lower bound of the input data interval
         * @param upper_bound Upper bound of the input data interval
         * @param size Number of input-output pairs generated
         * @param output Constant output value
         */
        LIB4NEURO_API DataSet(double lower_bound,
                              double upper_bound,
                              unsigned int size,
                              double output,
                              NormalizationStrategy* ns = nullptr);

        LIB4NEURO_API ~DataSet();

        /**
         *
         * @param bounds
         * @param no_elems_in_one_dim
         * @param output_func
         * @param output_dim
         */
        LIB4NEURO_API DataSet(std::vector<double>& bounds,
                              unsigned int no_elems_in_one_dim,
                              std::vector<double> (* output_func)(std::vector<double>&),
                              unsigned int output_dim,
                              NormalizationStrategy* ns = nullptr);

        /**
         *
         */
        LIB4NEURO_API void shift_outputs_to_zero();

        /**
         * Getter for number of elements
         * @return Number of elements in the data set
         */
        LIB4NEURO_API size_t get_n_elements() const ;

        /**
         * Returns the input dimension
         * @return Input dimension
         */
        LIB4NEURO_API size_t get_input_dim() const ;

        /**
         * Return the output dimension
         * @return Output dimension
         */
        LIB4NEURO_API size_t get_output_dim() const ;

        /**
         * Getter for the data structure
         * @return Vector of data
         */
        LIB4NEURO_API std::vector<std::pair<std::vector<double>, std::vector<double>>>* get_data();

        /**
         * Adds a new pair of data to the data set
         * @param inputs Vector of input data
         * @param outputs Vector of output data corresponding to the input data
         */
        LIB4NEURO_API void add_data_pair(std::vector<double>& inputs,
                                         std::vector<double>& outputs);

        //TODO expand method to generate multiple data types - chebyshev etc.
        /**
         * Adds a new data with input values equidistantly positioned
         * over the certain interval and the output value
         * being constant
         *
         * Both input and output are 1-dimensional
         *
         * @param lower_bound Lower bound of the input data interval
         * @param upper_bound Upper bound of the input data interval
         * @param size Number of input-output pairs generated
         * @param output Constant output value
         */
        LIB4NEURO_API void add_isotropic_data(double lower_bound,
                                              double upper_bound,
                                              unsigned int size,
                                              double output);

        /**
         * Adds a new data with input values equidistantly positioned
         * over the certain interval and the output value
         * being constant
         *
         * Input can have arbitrary many dimensions,
         * output can be an arbitrary function
         *
         * @param bounds Odd values are lower bounds and even values are corresponding upper bounds
         * @param size Number of input-output pairs generated
         * @param output_func Function determining output value
         */
        LIB4NEURO_API void add_isotropic_data(std::vector<double>& bounds,
                                              unsigned int no_elems_in_one_dim,
                                              std::vector<double> (* output_func)(std::vector<double>&));

        //TODO Chebyshev - ch. interpolation points, i-th point = cos(i*alpha) from 0 to pi

        /**
         * Prints the data set
         */
        LIB4NEURO_API void print_data();

        /**
         * Stores the DataSet object to the binary file
         *
         */
        LIB4NEURO_API void store_text(std::string file_path);

        /**
         *
         * @param file_path
         */
        LIB4NEURO_API void store_data_text(std::ofstream* file_path);

        /**
         * Stores the data to the text file in a human readable format
         *
         * @param file_path
         */
        LIB4NEURO_API void store_data_text(std::string file_path);

        /**
         * Normalizes the data set
         */
        LIB4NEURO_API void normalize();
		
		LIB4NEURO_API void normalize_inputs(
			const std::vector<double> &vals_min,
			const std::vector<double> &vals_max
		);

		LIB4NEURO_API void normalize_outputs(
			const std::vector<double> &vals_min,
			const std::vector<double> &vals_max
		);

        /**
         * returns the normalized value of @val
         * @param val
         * @return
         */
        LIB4NEURO_API double get_normalized_value(double val);

        /**
         *
         * @param val
         * @return
         */
        LIB4NEURO_API double get_denormalized_value(double val);

        /**
         * Denormalizes the data set
         */
        LIB4NEURO_API void de_normalize();

        /**
         * stores the de-normalized vector @d1 into @d2
         * @param d1
         * @param d2
         */
        LIB4NEURO_API void de_normalize_single(std::vector<double>& d1,
                                               std::vector<double>& d2);

        /**
         * stores the @idx-th input in the vector @d
         * @param d
         * @param idx
         */
        LIB4NEURO_API void get_input(std::vector<double>& d,
                                     size_t idx);

        /**
         * stores the @idx-th output in the vector @d
         * @param d
         * @param idx
         */
        LIB4NEURO_API void get_output(std::vector<double>& d,
                                      size_t idx);

        /**
         *
         * @return
         */
        LIB4NEURO_API NormalizationStrategy* get_normalization_strategy();

        LIB4NEURO_API void set_normalization_strategy(NormalizationStrategy* ns);

        /**
         *
         * @return
         */
        LIB4NEURO_API bool is_normalized();

        /**
         *
         * @return
         */
        LIB4NEURO_API double get_max_inp_val();

        /**
         *
         * @return
         */
        LIB4NEURO_API double get_min_inp_val();

        /**
         *
         * @param max
         * @return
         */
        [[deprecated("get_next_random_data_batch() should be used instead")]]
        LIB4NEURO_API std::vector<std::pair<std::vector<double>, std::vector<double>>>
        get_random_data_batch(size_t max);

        /**
         * Function, that returns vector with the next random disjunct batch of DataSet.
         * Randomness is given by copying the current DataSet, shuffling it once by Fisher-Yates algorithm and then
         * selecting first 'max' elements, which weren't selected before.
         *
         * If the shuffled vector is depleted, the process repeats with copying the DataSet again and another random
         * shuffle...
         *
         * @param max
         * @return
         */
        [[nodiscard]] LIB4NEURO_API std::vector<std::pair<std::vector<double>, std::vector<double>>> get_next_data_batch(size_t max);

        /**
         *
         * @param batch_size
         */
        LIB4NEURO_API void activate_next_data_batch(size_t batch_size);

		/**
		 * Adds a new output column filled with zeros
		 * @param n_columns Number of columns to be inserted
		 */
		LIB4NEURO_API void add_zero_output_columns(size_t n_columns);

        /*
        takes the data on all processes and makes it so that each MPI process
        has roughly the same number of data entries
        */
        LIB4NEURO_API void MPI_redistribute_data( );

        /*
        gather all data entries on the master process
        */
        LIB4NEURO_API void MPI_gather_data_on_master( );

        [[nodiscard]] LIB4NEURO_API arma::Mat<double>* get_inputs_matrix();

        [[nodiscard]] LIB4NEURO_API arma::Mat<double>* get_outputs_matrix();

        [[nodiscard]] LIB4NEURO_API arma::Mat<double>* get_stability_matrix();

        /*
        constructs an approximation of this DataSet. It groups together the entries according
        to the inputs, interpolates the output
        */
        [[nodiscard]] LIB4NEURO_API DataSet* get_approximated_set( double resolution, size_t power_ ) const;

        /*
        constructs a new dataset with altered outputs according to the parameter @resolution
        when @resolution = 0, then the outputs are transformed into a constant value
        when @resolution > 0, then the outputs correspond to the constant value with added difference towards
        the original outputs scaled by @resolution
        */
        [[nodiscard]] LIB4NEURO_API DataSet* get_smoothed_set( double resolution ) const;


        [[nodiscard]] LIB4NEURO_API std::vector<unsigned int>* get_active_elements_indices();

        [[nodiscard]] LIB4NEURO_API std::vector<unsigned int>* get_permutation_data_indices();

        [[nodiscard]] LIB4NEURO_API unsigned int* get_permutation_data_indices_content();

        [[nodiscard]] LIB4NEURO_API unsigned int* get_active_elements_content();

        [[nodiscard]] LIB4NEURO_API unsigned int get_current_batch_vector_index();

        [[nodiscard]] LIB4NEURO_API unsigned int* get_current_batch_vector_index_ptr();

        [[nodiscard]] LIB4NEURO_API unsigned int* get_num_active_elements_ptr();

    };
}
#endif //INC_4NEURO_DATASET_H
