//
// Created by martin on 05.09.19.
//

#define ARMA_ALLOW_FAKE_GCC
#include <armadillo>
#include <utility>
#include <random>

#include "../exceptions.h"
#include "../message.h"

#include "NelderMead.h"

lib4neuro::NelderMead::NelderMead(double reflectionCoefficient,
                                  double expansionCoefficient,
                                  double contractionCoefficient,
                                  double shrinkCoefficient,
                                  long long int maximumNiters,
                                  double terminating_net_error,
                                  double terminating_vertex_difference,
                                  double terminating_simplex_contraction,
                                  double initial_edge_size)
    : maximum_niters(maximumNiters), reflection_coefficient(reflectionCoefficient),
      expansion_coefficient(expansionCoefficient), contraction_coefficient(contractionCoefficient),
      shrink_coefficient(shrinkCoefficient), terminating_net_error(terminating_net_error),
      terminating_vertex_difference(terminating_vertex_difference),
      terminating_simplex_contraction(terminating_simplex_contraction),
      initial_edge_size(initial_edge_size) {

    if(reflectionCoefficient <= 0) {
        THROW_LOGIC_ERROR("Reflection coefficient must be greater than 0!");
    }

    if(expansionCoefficient <=1 ) {
        THROW_LOGIC_ERROR("Expansion coefficient must be greater than 1!");
    }

    if(! (contractionCoefficient > 0 && contractionCoefficient <1 )) {
        THROW_LOGIC_ERROR("Contraction coefficient must be between 0 and 1!");
    }

    if(! (shrinkCoefficient > 0 && shrinkCoefficient < 1)) {
        THROW_LOGIC_ERROR("Shrink coefficient must be between 0 and 1!");
    }
}

lib4neuro::NelderMead::NelderMead(long long int maximumNiters,
                                  double initial_edge_size)
                                  : NelderMead(1,
                                               2,
                                               0.5,
                                               0.5,
                                               maximumNiters,
                                               1e-4,
                                               1e-5,
                                               1e-5,
                                               initial_edge_size) {}

lib4neuro::NelderMead::NelderMead(double initial_edge_size) : NelderMead(1,
                                                                         2,
                                                                         0.5,
                                                                         0.5,
                                                                         500,
                                                                         1e-5,
                                                                         1e-5,
                                                                         1e-5,
                                                                         initial_edge_size) {}

long long int lib4neuro::NelderMead::getMaximumNiters() const {
    return maximum_niters;
}

double lib4neuro::NelderMead::getReflectionCoefficient() const {
    return reflection_coefficient;
}

double lib4neuro::NelderMead::getExpansionCoefficient() const {
    return expansion_coefficient;
}

double lib4neuro::NelderMead::getContractionCoefficient() const {
    return contraction_coefficient;
}

double lib4neuro::NelderMead::getShrinkCoefficient() const {
    return shrink_coefficient;
}

lib4neuro::NelderMead::~NelderMead() {}

void lib4neuro::NelderMead::optimize(lib4neuro::ErrorFunction& ef,
                                     std::ofstream* ofs) {

    double err_ = ef.eval();
    COUT_INFO("Nelder-Mead optimizer starting...");
    COUT_INFO("Initial network error: " + std::to_string(err_));
    size_t problem_dim = ef.get_dimension();
    this->optimized_function_dimension = problem_dim;
    unsigned int best = 0;
    unsigned int worst = problem_dim;
    unsigned int second_worst = problem_dim-1;

    /* Initialize simplex */
    /* Every column corresponds to a coordinates of one point in a simplex */
    this->simplex = new arma::Mat<double>(problem_dim, problem_dim + 1);

    /* Initialize first point coordinates with network's parameters */
    arma::Col<double> column(ef.get_parameters());
    this->simplex->col(0) = column;

    for(size_t i = 1; i <= problem_dim; i++) {
        this->simplex->col(i) = this->simplex->col(i-1);
        this->simplex->col(i)(i-1) += this->initial_edge_size;
    }

    /* Function values at simplex's points */
    /* Single output is a single column */
    arma::Col<double> simplex_function_values(problem_dim+1);
    std::vector<double> std_vec_tmp;

    /* Indices of a sorted simplex points, according to their function values */
    arma::Col<arma::uword> simplex_indices;

    std::vector<double> init_net_parameters = ef.get_parameters();

    this->optimal_parameters = ef.get_parameters();

    for(unsigned int iter_ind = 0; iter_ind < this->maximum_niters; iter_ind++) {
        /* Eval function */
        for(size_t j = 0; j <= problem_dim; j++) {
            std_vec_tmp = arma::conv_to<std::vector<double>>::from(this->simplex->col(j));
            simplex_function_values(j) = ef.eval(&std_vec_tmp);  //TODO do in a more efficient way
        }

        /* Sort simplex points according to the error function output */
        simplex_indices = arma::sort_index(simplex_function_values);
        *this->simplex = this->simplex->cols(simplex_indices);
        simplex_function_values = arma::sort(simplex_function_values);

        this->optimal_parameters = arma::conv_to<std::vector<double>>::from(this->simplex->col(best));
        ef.set_parameters(this->optimal_parameters);

        /* Termination conditions (network error,  */
        err_ = ef.eval();
        if(err_ < this->terminating_net_error
           || std::abs(simplex_function_values.at(best) - simplex_function_values.at(worst)) < this->terminating_vertex_difference
           || this->get_simplex_contraction() < this->terminating_simplex_contraction) {
            COUT_INFO("Nelder-Mead optimizer converged after " + std::to_string(iter_ind+1) + " iterations"
                      " with the final error " + std::to_string(err_) + ".");

            return;
        }

        /* Compute centroid */
        arma::Col<double> centroid(problem_dim);
        for(size_t j = 0; j < problem_dim; j++) {
            centroid(j) = arma::mean(this->simplex->row(j).subvec(0, problem_dim-1));
        }

        /* Reflection */
        arma::Col<double> xr = centroid + this->reflection_coefficient*(centroid - this->simplex->col(worst));  // Reflection point
        std_vec_tmp = arma::conv_to<std::vector<double>>::from(xr);
        double xr_val = ef.eval(&std_vec_tmp);
        ef.set_parameters(this->optimal_parameters);

        if(xr_val < simplex_function_values(best) ) {
            /* Expansion */
            arma::Col<double> xe = centroid + this->expansion_coefficient*(xr - centroid);
            std_vec_tmp = arma::conv_to<std::vector<double>>::from(xe);
            double xe_val = ef.eval(&std_vec_tmp);
            ef.set_parameters(this->optimal_parameters);

            if(xe_val < xr_val) {
//                std::cout << simplex_function_values(problem_dim) << " " << xe_val << std::endl;
                this->simplex->col(worst) = xe;
                continue;
            } else {
//                std::cout << simplex_function_values(problem_dim) << " " << xr_val << std::endl;
                this->simplex->col(worst) = xr;
                continue;
            }

        } else if(xr_val < simplex_function_values(second_worst)) {
            /* Reflection */
            this->simplex->col(worst) = xr;
            continue;

        } else {
            /* Contraction */
            arma::Col<double> xc = centroid + this->contraction_coefficient*(this->simplex->col(worst) - centroid);
            std_vec_tmp = arma::conv_to<std::vector<double>>::from(xc);
            double xc_value = ef.eval(&std_vec_tmp);
            ef.set_parameters(this->optimal_parameters);

            if(xc_value < simplex_function_values(worst)) {
                this->simplex->col(worst) = xc;
                continue;
            } else {
                /* Shrink */
                for(size_t j = 0; j <= problem_dim; j++) {
                    if(j == best) {
                        continue;
                    }

                    this->simplex->col(j) = centroid + this->shrink_coefficient*(this->simplex->col(j) - this->simplex->col(best));
                }
            }
        }
    }

    this->optimal_parameters = arma::conv_to<std::vector<double>>::from(this->simplex->col(best));
    ef.set_parameters(this->optimal_parameters);

    err_ = ef.eval();
    COUT_INFO("Maximal number of iterations (" + std::to_string(this->maximum_niters) + ") was reached! Network "
                   "error is " + std::to_string(err_) + ".");
}

double lib4neuro::NelderMead::get_simplex_contraction() {
    size_t dim = this->optimized_function_dimension;
    arma::Col<double> distances(dim);

    for(size_t i = 0; i < dim; i++) {
        distances.at(i) = arma::norm(this->simplex->col(0) - this->simplex->col(i+1));
    }

    arma::Col<double> denominator_vec = {1, arma::norm(this->simplex->col(0))};
    return arma::max(distances)/arma::max(denominator_vec);
}

double lib4neuro::NelderMead::getTerminatingNetError() const {
    return terminating_net_error;
}

double lib4neuro::NelderMead::getTerminatingVertexDifference() const {
    return terminating_vertex_difference;
}

double lib4neuro::NelderMead::getTerminatingSimplexContraction() const {
    return terminating_simplex_contraction;
}
