#include <string>
#include <fstream>
#include <sstream>
#include <regex>
#include <algorithm>
#include <memory>
#include <boost/lexical_cast.hpp>
#include <boost/algorithm/string/erase.hpp>
#include <mpi.h>

#include "../mpi_wrapper.h"

#include "CSVReader.h"
#include "exceptions.h"

bool is_file_accessible(std::string file_path) {
    return std::ifstream(file_path).good();
}

namespace lib4neuro {
    CSVReader::CSVReader(std::string file_path,
                         std::string delimiter,
                         bool ignore_first_line) : Reader(file_path, ignore_first_line, delimiter) {
        if( lib4neuro::mpi_rank == 0 ){
            if (!is_file_accessible(file_path)) {
                MPI_INTERRUPT
                THROW_RUNTIME_ERROR("The file path \'" + file_path + "\' specified in Reader is not accessible!");
            }
            this->header_included   = ignore_first_line;
        }
        MPI_ERROR_CHECK
    }

    std::shared_ptr<DataSet> CSVReader::get_data_set(std::vector<unsigned int>* input_col_indices,
                                                     std::vector<unsigned int>* output_col_indices) {

        //TODO reimplement as lazy loading using data_set variable inherited from Reader class
        std::vector<std::pair<std::vector<double>, std::vector<double>>> data_set_contents;
        int inp_dim = 0;
        int out_dim = 0;
        if( lib4neuro::mpi_rank == 0 ){

            if (this->data.empty()) {
                MPI_INTERRUPT
                THROW_LOGIC_ERROR("DataSet can not be created as there were no data read beforehand! Did you forget to call "
                                  "the method 'read()'?");
            }

            for (auto line : this->data) {
                //TODO check empty values in data
                std::vector<double> input;
                for (auto           ind : *input_col_indices) {
                    std::string s;

                    try {
                        /* Remove remaining spaces */
                        s = line.at(ind);
                        boost::algorithm::erase_all(s,
                                                    " ");

                        /* Strip BOM */
                        // TODO solve in another way - work properly with different encodings!
                        boost::algorithm::erase_all(s,
                                                    "\uEFBBBF");  // UTF-8
                        boost::algorithm::erase_all(s,
                                                    "\uFEFF");  // UTF-16

                        /* Check, if the string is a number */
                        auto tmp = boost::lexical_cast<double>(s);

                        /* Add loaded number to the vector of inputs */
                        input.push_back(tmp);

                    }
                    catch (const std::out_of_range& e) {
                        MPI_INTERRUPT
                        THROW_OUT_OF_RANGE_ERROR("Non-existing index specified (" + std::to_string(ind) + ")!");

                    }
                    catch (const boost::bad_lexical_cast& e) {
                        MPI_INTERRUPT
                        THROW_RUNTIME_ERROR(
                            "Value \"" + s + "\" is not numerical and so it cannot be used in Data Set!");
                    }
                }

                std::vector<double> output;
                for (auto           ind : *output_col_indices) {
                    output.emplace_back(std::stod(line.at(ind)));
                }

                data_set_contents.emplace_back(std::make_pair(input,
                                                              output));
                inp_dim = input.size();
                out_dim = output.size();
            }
        }
        MPI_ERROR_CHECK

        MPI_Bcast( &inp_dim, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
        MPI_Bcast( &out_dim, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );

        if( lib4neuro::mpi_rank == 0 ){
            auto out = std::make_shared<DataSet>(DataSet(&data_set_contents));
            out->MPI_redistribute_data();
            return out;
        }
        else{
            auto out = std::make_shared<DataSet>(DataSet( inp_dim, out_dim ));
            out->MPI_redistribute_data();
            return out;
        }

    }
}
