//
// Created by martin on 16.08.19.
//

#include <armadillo>
#include <limits>
#include <mpi.h>

#include "../mpi_wrapper.h"
#include "../message.h"

#include "XYZReader.h"
#include "../exceptions.h"

lib4neuro::XYZReader::XYZReader(const std::string& file_path, bool ignore_charge) : Reader(file_path,
                                                                             false,
                                                                                  " ") {
    this->element_list = std::make_shared<std::vector<ELEMENT_SYMBOL>>(std::vector<ELEMENT_SYMBOL>());
    this->acsf_data_set = nullptr;
    this->charge = false;
    this->ignore_charge = ignore_charge;
}

std::shared_ptr<lib4neuro::DataSet> lib4neuro::XYZReader::get_data_set() {

    std::vector<std::pair<std::vector<double>, std::vector<double>>> data_set_contents;
    int inp_dim = 0;
    int out_dim = 0;
    int n_clusters = 0;
    int n_elements = 0;
    if( lib4neuro::mpi_rank == 0 ){
        if(this->data_set == nullptr) {


            if (this->data.empty()) {
                MPI_INTERRUPT
                THROW_LOGIC_ERROR("DataSet can not be created as there were no data read beforehand! Did you forget to call "
                                  "the method 'read()'?");
            }

            unsigned int n_configurations;
            unsigned int n_particles;

            std::vector<double> inputs, outputs;

            /* Read number of configurations */
            std::vector<std::string> line = this->data.at(0);
            this->remove_white_characters_from_vector(line);
            n_configurations = std::stoul(line.at(0));
            data_set_contents.reserve(n_configurations);

            /* Read number of particles per single configuration */
            line = this->data.at(1);
            this->remove_white_characters_from_vector(line);
            n_particles = std::stoul(line.at(0));
            // COUT_INFO( "# of particles in cluster: " << n_particles);
//            COUT_INFO( "# of various particles: " << n_configurations);

            unsigned int situation;

            for (size_t data_ind = 2; data_ind < this->data.size(); data_ind++) {
                /* Read data for a new line */
                line = this->data.at(data_ind);

                /* Remove empty strings from the line vector left after "split */
                this->remove_white_characters_from_vector(line);

                situation = data_ind % (n_particles + 2);
                if (situation == 1) {
                    data_set_contents.emplace_back(std::make_pair(inputs,
                                                                  outputs));
    //                std::cout << "[" << data_set_contents.size() << "] inputs: ";
    //                for( auto el: inputs ){
    //                    std::cout << el << ", ";
    //                }
    //                std::cout << "outputs: ";
    //                for( auto el: outputs ){
    //                    std::cout << el << ", ";
    //                }
    //                std::cout << std::endl;

                    inputs.clear();
                    outputs.clear();
                } else if (situation == 2) {
                    /* Save energy as a network output */
                    outputs = {std::stod(line.at(0))};
                    if(line.size() == 2) {
                        /* Ignore data about added temperature */
                        // TODO make more flexible - we can use it sometimes!
                        outputs = {std::stod(line.at(1))};
                    }

                } else {
                    /* Save element symbol */
                    if(this->element_list->size() < n_particles) {
                        this->element_list->emplace_back(this->get_element_symbol(line.at(0)));
                    }

                    /* Save coordinates */
                    for (size_t i = 1; i < line.size(); i++) {
                        inputs.emplace_back(std::stod(line.at(i)));
                    }
                }
            }
            data_set_contents.emplace_back(std::make_pair(inputs, outputs));
            inp_dim = inputs.size();
            out_dim = outputs.size();

            inputs.clear();
            outputs.clear();

            /* Compute number of particles per element */
            for(auto e : *this->element_list) {
                if( this->n_particles_per_element[e] ) {
                    this->n_particles_per_element[e]++;
                } else {
                    this->n_particles_per_element[e] = 1;
                }
            }
            /* Is charge included in data? */
            std::vector<std::string> tmp = this->data.at(3);
            this->remove_white_characters_from_vector(tmp);


            this->charge = (tmp.size() == 5) ? true : false;

            n_clusters = data_set_contents.size( );
            n_elements = this->element_list->size( );
        }
    }
    MPI_ERROR_CHECK

    MPI_Bcast( &inp_dim, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
    MPI_Bcast( &out_dim, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
    MPI_Bcast( &n_clusters, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
    MPI_Bcast( &n_elements, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
    MPI_Bcast( &this->charge, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
    // COUT_INFO("# number of configurations: " << n_clusters);

    if( this->data_set == nullptr ){
        /* synchronization of element list */
        if( lib4neuro::mpi_rank == 0 ){
            for( int i = 0; i < n_elements; ++i ){
                int particle_idx = -1;
                if(this->element_list->at(i) == ELEMENT_SYMBOL::He){
                    particle_idx = 0;
                }
                else if(this->element_list->at(i) == ELEMENT_SYMBOL::N){
                    particle_idx = 1;
                }
                MPI_Bcast( &particle_idx, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );
            }
        }
        else{
            for( int i = 0; i < n_elements; ++i ){
                int particle_idx;
                MPI_Bcast( &particle_idx, 1, MPI_INT, 0, lib4neuro::mpi_active_comm );

                if( particle_idx == 0 ){
                    this->element_list->emplace_back(ELEMENT_SYMBOL::He);
                }
                else if( particle_idx == 1 ){
                    this->element_list->emplace_back(ELEMENT_SYMBOL::N);
                }
            }
        }

        /* data set creation */
        if( lib4neuro::mpi_rank == 0 ){
            this->data_set = std::make_shared<DataSet>(DataSet(&data_set_contents));
        }
        else{
            this->data_set = std::make_shared<DataSet>(DataSet( inp_dim, out_dim ));
        }
        this->data_set->MPI_redistribute_data();
    }

    return this->data_set;
}

std::shared_ptr<std::vector<lib4neuro::ELEMENT_SYMBOL>> lib4neuro::XYZReader::get_element_list() {
    if(this->element_list == nullptr) {
        this->element_list = std::shared_ptr<std::vector<ELEMENT_SYMBOL>>(this->element_list);
    }
    return this->element_list;
}

void lib4neuro::XYZReader::transform_input_to_acsf(std::unordered_map<ELEMENT_SYMBOL, Element*>& element_description) {
    //TODO implement a separate class CoordinatesTransformator and just call its method here

    std::vector<std::pair<std::vector<double>, std::vector<double>>>* data = this->get_data_set()->get_data();
	size_t  out_dim = 0;
	if( data->size() > 0 ){
		out_dim = data->at( 0 ).second.size();
	}

    unsigned int n_particles = this->element_list->size();
    unsigned int n_coords_per_particle = (this->charge && !this->ignore_charge) ? 4 : 3;

	COUT_INFO("data contains charge " << this->charge );
	COUT_INFO("should we ignore charge " << this->ignore_charge );

    /* Create a new set of Atomic-Centered Symmetry Functions serving as coordinates */
    std::vector<double> acsf_coords;

    std::vector<std::pair<ELEMENT_SYMBOL, std::vector<double>>> particles;
    particles.reserve(n_particles);
    std::vector<double> cartesian_coords;
    cartesian_coords.resize(3);

    unsigned int   idx;
    this->acsf_data_set = std::make_shared<DataSet>();
	
	/* normalization data for each element*/
	std::unordered_map<ELEMENT_SYMBOL, std::vector<double> > inputs_min, inputs_max;
	std::vector<double> outputs_min(out_dim), outputs_max(out_dim);
	std::fill(outputs_min.begin(), outputs_min.end(), std::numeric_limits<double>::max());
	std::fill(outputs_max.begin(), outputs_max.end(), std::numeric_limits<double>::min());
	for( auto el: element_description ){
		inputs_min[el.first] = std::vector<double>(element_description[el.first]->getSymmetryFunctions()->size());
		inputs_max[el.first] = std::vector<double>(element_description[el.first]->getSymmetryFunctions()->size());
		std::fill(inputs_min[el.first].begin(), inputs_min[el.first].end(), std::numeric_limits<double>::max());
		std::fill(inputs_max[el.first].begin(), inputs_max[el.first].end(), std::numeric_limits<double>::min());
		
		// std::cout << "new element: " << inputs_min[el.first].size() << std::endl;
	}
	
	std::vector<ELEMENT_SYMBOL> element_order;//used for input ordering

    for(auto       configuration : *data) { /* Iterate over configurations */
	
		// std::cout << "A" << std::endl;
		/* update the output value ranges */
		for( size_t j = 0; j < outputs_min.size(); ++j ){
			outputs_min[ j ] = std::min( outputs_min[ j ], configuration.second[ j ] );
			outputs_max[ j ] = std::max( outputs_max[ j ], configuration.second[ j ] );
		}
		// std::cout << "B" << std::endl;
        idx = 0;
        particles.clear();

        /* Save Cartesian coordinates into the vector */
        for(auto el : *this->element_list) { /* Iterate over particles */
            for(unsigned int i = 0; i < 3; i++) { /* Iterate over all coordinates of a single particle */
                cartesian_coords.at(i) = configuration.first.at(idx);
//				std::cout <<  "d[" << i << "] = i[" << idx << "](" << configuration.first.at(idx) << "), " ;
                idx++;
            }
			if( this->charge && !this->ignore_charge ){
                cartesian_coords.at(3) = configuration.first.at(idx);
//				std::cout <<  "d[" << 3 << "] = i[" << idx << "](" << configuration.first.at(idx) << ") " ;
                idx++;
			}
//			std::cout << std::endl;

            if(this->ignore_charge && this->charge) {
                idx++; //skips charge - TODO add support for charge!
            }

            particles.emplace_back(std::make_pair(el, cartesian_coords));
        }
		// std::cout << "C" << std::endl;

        /* Transform Cartesian coordinates into ACSFs */
        std::vector<double> single_particle_coords;
        std::vector<std::pair<ELEMENT_SYMBOL, std::vector<double>>> acsf_coords_prep;
        for(size_t i = 0; i < particles.size(); i++) { /* Iterate over all the particles */
            single_particle_coords.clear();
			// std::cout << "D" << std::endl;

            /* Compute vector of ACSF coordinates for a single particle */
            for (auto sym_func : *element_description[particles.at(i).first]->getSymmetryFunctions()) {
				// std::cout << "E1" << std::endl;
                single_particle_coords.emplace_back(sym_func->eval(i, particles));
				// std::cout << "E2" << std::endl;
            }

			// std::cout << "X" << std::endl;
			/* update the input value ranges */
			for( auto j = 0; j < single_particle_coords.size(); ++j ){
				inputs_min[particles.at(i).first][ j ] = std::min(inputs_min[particles.at(i).first][ j ], single_particle_coords[ j ]);
				inputs_max[particles.at(i).first][ j ] = std::max(inputs_max[particles.at(i).first][ j ], single_particle_coords[ j ]);
			}
			// std::cout << "Y" << std::endl;

            /* Store coordinates in a canonic order */
            unsigned int element_ind = 0;
            if(acsf_coords_prep.empty()) {
                acsf_coords_prep.emplace_back(std::make_pair(particles.at(i).first, single_particle_coords));
            } else {
                /* Find a position to insert the element coordinates according to the periodic table */
                for(size_t j = 0; j < acsf_coords_prep.size(); j++) {
                    element_ind = j+1;
                    if(acsf_coords_prep.at(j).first > particles.at(i).first) {
                        break;
                    }
                }

                /* Find a position to insert the element coordinates according to the coordinates' vector norm */
                while(element_ind > 0) {
                    if(acsf_coords_prep.at(element_ind-1).first != particles.at(i).first) {
                        break;
                    }

                    if(arma::norm(arma::conv_to<arma::Col<double>>::from(single_particle_coords)) >
                        arma::norm(arma::conv_to<arma::Col<double>>::from(acsf_coords_prep.at(element_ind-1).second))) {
                        break;
                    }

                    element_ind--;
                }

                acsf_coords_prep.insert(acsf_coords_prep.begin()+element_ind,
                                        std::make_pair(particles.at(i).first, single_particle_coords));
            }
        }
		
		if( element_order.size() == 0 ){
			for(auto e : acsf_coords_prep) {
				element_order.push_back(e.first);
			}
		}

        /* Rewrite coordinates without ELEMENT_SYMBOL labels */
        for(auto e : acsf_coords_prep) {
            for(auto coord : e.second) {
                acsf_coords.emplace_back(coord);
            }
        }

        /* Check for duplicates in new coordinates */
//        std::vector<std::vector<double>> unique_coords = coords_check;
//        std::sort(unique_coords.begin(), unique_coords.end());
//        unique_coords.erase(std::unique(unique_coords.begin(), unique_coords.end()), unique_coords.end());
//
//        if(unique_coords.size() != coords_check.size()) {
//            MPI_INTERRUPT
//            THROW_RUNTIME_ERROR("Not all descriptors are unique with currently specified symmetry functions!");
//        }

        this->acsf_data_set->add_data_pair(acsf_coords, configuration.second);
        acsf_coords.clear();
        acsf_coords_prep.clear();
    }
	
	/* sycnhronization of normalized data */
	MPI_Allreduce( MPI_IN_PLACE, &outputs_min[ 0 ], outputs_min.size(), MPI_DOUBLE, MPI_MIN, lib4neuro::mpi_active_comm );
	MPI_Allreduce( MPI_IN_PLACE, &outputs_max[ 0 ], outputs_max.size(), MPI_DOUBLE, MPI_MAX, lib4neuro::mpi_active_comm );
	for( auto el: element_description ){
		MPI_Allreduce( MPI_IN_PLACE, &inputs_min[el.first][ 0 ], inputs_min[el.first].size(), MPI_DOUBLE, MPI_MIN, lib4neuro::mpi_active_comm );
		MPI_Allreduce( MPI_IN_PLACE, &inputs_max[el.first][ 0 ], inputs_max[el.first].size(), MPI_DOUBLE, MPI_MAX, lib4neuro::mpi_active_comm );
	}
	
	std::vector<double> input_normalization_min, output_normalization_min;
	std::vector<double> input_normalization_max, output_normalization_max;
	for( auto el: element_order ){
		for( auto val: inputs_min[el] ){
			input_normalization_min.push_back( val );
		}
		for( auto val: inputs_max[el] ){
			input_normalization_max.push_back( val );
		}
	}
	output_normalization_min = outputs_min;
	output_normalization_max = outputs_max;
	
	
	if( lib4neuro::mpi_rank == 0 ){
		std::cout << "input ranges" << std::endl;
		for( size_t j = 0; j < input_normalization_min.size(); ++j ){
			std::cout << " " << input_normalization_min[j] << " - " << input_normalization_max[j] << std::endl;
		}
		std::cout << "output ranges" << std::endl;
		for( size_t j = 0; j < output_normalization_min.size(); ++j ){
			std::cout << " " << output_normalization_min[j] << " - " << output_normalization_max[j] << std::endl;
		}
	}
	
	this->acsf_data_set->normalize_inputs( input_normalization_min, input_normalization_max );
	this->acsf_data_set->normalize_outputs( output_normalization_min, output_normalization_max );
    MPI_ERROR_CHECK
}

std::shared_ptr<lib4neuro::DataSet> lib4neuro::XYZReader::get_acsf_data_set() {
    if(acsf_data_set == nullptr) {
        MPI_INTERRUPT
        THROW_LOGIC_ERROR("No ACSF data set was found!");
    }
    MPI_ERROR_CHECK
    return this->acsf_data_set;
}

lib4neuro::ELEMENT_SYMBOL lib4neuro::XYZReader::get_element_symbol(std::string symbol) {
    //TODO implement in a more elegant way

    std::unordered_map<std::string, ELEMENT_SYMBOL> element_symbols;
    element_symbols["He"] = ELEMENT_SYMBOL::He;
    element_symbols["N"] = ELEMENT_SYMBOL::N;

    return element_symbols[symbol];
}

std::shared_ptr<lib4neuro::DataSet>
lib4neuro::XYZReader::get_acsf_data_set(std::unordered_map<ELEMENT_SYMBOL, Element*>& element_description) {
    this->transform_input_to_acsf(element_description);
    return this->acsf_data_set;
}

bool lib4neuro::XYZReader::contains_charge() {
    return this->charge;
}

bool lib4neuro::XYZReader::get_ignore_charge() {
    return this->ignore_charge;
}
