/**
 * File containing methods for quick & simple formulation of PDEs as a system of Neural Networks with shared weights
 *
 * @author Michal Kravčenko
 * @date 22.7.18 -
 */

//TODO incorporate uncertainities as coefficients in NeuralNetworkSum or ErrorSum
//TODO add support for multiple unknown functions to be found
//TODO add export capability?
//TODO restructure of the learning methods to have a common parent to be used as a parameter in the solvers

#ifndef INC_4NEURO_PDESOLVER_H
#define INC_4NEURO_PDESOLVER_H

#include <map>

#include "../settings.h"
#include "../DataSet/DataSet.h"
#include "../Network/NeuralNetwork.h"
#include "../Network/NeuralNetworkSum.h"
#include "../ErrorFunction/ErrorFunctions.h"
#include "../Neuron/NeuronLinear.h"
#include "../Neuron/NeuronLogistic.h"
#include "../LearningMethods/ParticleSwarm.h"
#include "../LearningMethods/GradientDescent.h"

namespace lib4neuro {

/**
 * class representing a multi-index of partial derivatives
 */
    class MultiIndex {
    private:
        /**
         * total number of variables
         */
        size_t dim;

        /**
         * a vector containing degrees of partial derivatives with respect to each variable
         */
        std::vector<size_t> partial_derivatives_degrees;

    public:
        /**
         *
         * @param dimension
         */
        LIB4NEURO_API MultiIndex(size_t dimension);


        /**
         *
         * @param index
         * @param value
         */
        LIB4NEURO_API void set_partial_derivative(size_t index,
                                                  size_t value);

        /**
         *
         * @return
         */
        LIB4NEURO_API std::vector<size_t>* get_partial_derivatives_degrees();


        /**
         *
         * @param rhs
         * @return
         */
        LIB4NEURO_API bool operator<(const MultiIndex& rhs) const;

        /**
         *
         * @return
         */
        LIB4NEURO_API std::string to_string() const;

        /**
         *
         * @return
         */
        LIB4NEURO_API size_t get_degree() const;
    };


    class DESolver {
    private:

        /* Mapping between used multiindices of partial derivatices and already defined NNs (in order to not define
         * the same NN multiple times )*/
        std::map<MultiIndex, std::shared_ptr<NeuralNetwork>> map_multiindices2nn;

        /* A list of the differential equations */
        std::vector<std::shared_ptr<NeuralNetworkSum>> differential_equations; // = nullptr;

        /* Error functions for differential equations */
        std::vector<ErrorFunctionType> errors_functions_types; // = nullptr;
        std::vector<DataSet>           errors_functions_data_sets; // = nullptr;

        /* NN as the unknown function */
        std::shared_ptr<NeuralNetwork> solution            = std::make_shared<NeuralNetwork>(NeuralNetwork());

        /* auxilliary variables */
        std::vector<std::shared_ptr<NeuronLogistic>> solution_inner_neurons; // = nullptr;
        size_t                                       dim_i = 0, dim_inn = 0, n_equations = 0;

    public:
        /**
         * The attempted solution will contain 'm' inner neurons between the input neurons and the output neuron
         * @param n_equations
         * @param n_inputs
         * @param m
         */
        LIB4NEURO_API DESolver(size_t n_equations,
                               size_t n_inputs,
                               size_t m);

        /**
         * default destructor
         */
        LIB4NEURO_API ~DESolver();

        /**
         * Adds a new summand multiplied by 'beta' into the 'equation_idx'-th differential equation
         * @param equation_idx
         * @param alpha
         * @param beta
         */
        LIB4NEURO_API void
        add_to_differential_equation(size_t equation_idx,
                                     MultiIndex& alpha,
                                     std::string expression_string);


        /**
         * Adds a known function to the equation
         * @param equation_idx
         * @param expression_string
         */
        LIB4NEURO_API void add_to_differential_equation(size_t equation_idx,
                                                        std::string expression_string);

        /**
         * Sets the error function for the differential equation with the corresponding index
         * @param equation_idx
         * @param F
         * @param conditions
         */
        LIB4NEURO_API void set_error_function(size_t equation_idx,
                                              ErrorFunctionType F,
                                              DataSet& conditions);

        /**
         *
         * @param learning_method
         */
        LIB4NEURO_API void solve(LearningMethod& learning_method);

        /**
         *
         */
        LIB4NEURO_API void randomize_parameters();

        /**
         * returns the pointer to the object representing the given partial derivative of the solution
         * @return
         */
        LIB4NEURO_API NeuralNetwork* get_solution(MultiIndex& alpha);

        /**
         * For testing purposes only
         */
        LIB4NEURO_API double
        eval_equation(size_t equation_idx,
                      std::shared_ptr<std::vector<double>> weights_and_biases,
                      std::vector<double>& input);

        /**
         * For testing purposes only
         * @return
         */
        LIB4NEURO_API double eval_total_error(std::vector<double>& weights_and_biases);
    };

}
#endif //INC_4NEURO_PDESOLVER_H
