//
// Created by martin on 10.09.19.
//
#include <algorithm>
#include <random>
#include <unordered_map>

#include <armadillo>
#include <unordered_set>

#include <boost/functional/hash.hpp>
#include <mpi.h>

#include "../mpi_wrapper.h"

#include "LearningMethods/LevenbergMarquardt.h"
#include "Reader/XYZReader.h"
#include "../exceptions.h"
#include "../message.h"
#include "ACSFParametersOptimizer.h"
#include "../Network/ACSFNeuralNetwork.h"
#include "../LearningMethods/GradientDescentBB.h"

lib4neuro::ACSFParametersOptimizer::ACSFParametersOptimizer(lib4neuro::ErrorFunction* ef, XYZReader* reader) {
    this->error_func = ef;

    if(this->error_func->get_nets().size() > 1) {
        THROW_LOGIC_ERROR("Only single network can be passed as an argument!");
    }

    this->net = dynamic_cast<ACSFNeuralNetwork*>(this->error_func->get_nets().at(0));
    this->data_set = this->error_func->get_dataset();
    this->reader = reader;
}

void lib4neuro::ACSFParametersOptimizer::fit_ACSF_parameters(std::vector<SYMMETRY_FUNCTION_PARAMETER>& fitted_params,
                                                             bool random_init_params) {
    fit_ACSF_parameters(fitted_params,
                        random_init_params,
                        50,
                        10,
                        1e-5,
                        0.98,
                        2,//0.085,
                        1e-5);
}

void lib4neuro::ACSFParametersOptimizer::fit_ACSF_parameters(std::vector<SYMMETRY_FUNCTION_PARAMETER>& fitted_params,
                                                             bool random_init_params,
                                                             unsigned int main_iterations_num,
                                                             unsigned int steady_temperature_iters_num,
                                                             double init_temperature,
                                                             double temperature_change,
                                                             double distribution_stddev,
                                                             double termination_threshold) {
    COUT_INFO("Starting optimizing of ACSF parameters...");

    /* Seed the random numbers' generator */
    std::random_device rd{};
    std::mt19937 gen{rd()};

    /* Construct and train a new linear neural network to evaluate the chosen symmetry functions' parameters */
    std::vector<unsigned int> neuron_numbers = {static_cast<unsigned int>(this->net->get_input_neurons_number()), 1};
    std::vector<NEURON_TYPE> type_hidden_neurons = {};
    FullyConnectedFFN nn(&neuron_numbers, &type_hidden_neurons);

    /* Initial parameters' values */
    // TODO take care of parameters' boundaries
    std::unordered_map<SYMMETRY_FUNCTION_PARAMETER, double> init_param_values;
    std::normal_distribution<> d{1, 0.085};  // TODO make probability distribution choice more flexible by adding function parameters
    std::uniform_real_distribution<> d_uni(0, 2);
    std::vector<short int> shift_max_vals = {-1, 1};
    if(random_init_params) {
        /* Iterate over element's description */
        for(auto el : *this->net->get_elements()) {
            for(auto func : *el.second->getSymmetryFunctions()) {
                for (auto param : *func->get_parameters()) {
                    /* Check, if the parameter is supposed to be optimized */
                    if (std::find(fitted_params.begin(),
                                  fitted_params.end(),
                                  param.first) != fitted_params.end()) {
                        if (param.first == SYMMETRY_FUNCTION_PARAMETER::SHIFT_MAX) {

                            float val = d_uni(gen);
                            MPI_Bcast( &val, 1, MPI_FLOAT, 0, lib4neuro::mpi_active_comm);
                            func->set_parameter(param.first,
                                                shift_max_vals.at((int)(val)));
                        } else {
                            float val = d(gen);
                            MPI_Bcast( &val, 1, MPI_FLOAT, 0, lib4neuro::mpi_active_comm);
                            func->set_parameter(param.first,
                                                d(gen));
                        }
                    }
                }
            }
        }

        this->data_set = this->reader->get_acsf_data_set(*this->net->get_elements()).get();
    }

    double previous_error;
    double current_error;
    double delta_error;

    double temperature = init_temperature;  //TODO make flexible by adding function parameters

    MSE mse_init(&nn, this->data_set);
    previous_error = mse_init.eval();
    MPI_Allreduce( MPI_IN_PLACE, &previous_error, 1, MPI_DOUBLE, MPI_SUM, lib4neuro::mpi_active_comm );

    /* Iteratively improve the parameters */
    unsigned int iter_ind = 0;
    for(unsigned int i = 0; i < main_iterations_num; i++) {
        /* Temperature iterations */
        for(unsigned int j = 0; j < steady_temperature_iters_num; j++) {
            /* Change the selected parameters in ACSFs */
            for (auto& el : *this->net->get_elements()) {
                for (auto& func : *el.second->getSymmetryFunctions()) {
                    for (auto& param : *func->get_parameters()) {
                        /* Check, if the parameter is supposed to be optimized */
                        if (std::find(fitted_params.begin(),
                                      fitted_params.end(),
                                      param.first) != fitted_params.end()) {
                            if (param.first == SYMMETRY_FUNCTION_PARAMETER::SHIFT_MAX) {
                                float val = d_uni(gen);
                                MPI_Bcast( &val, 1, MPI_FLOAT, 0, lib4neuro::mpi_active_comm);
                                func->set_parameter(param.first,
                                                    shift_max_vals.at((int) (val)));
                            } else {
                                std::normal_distribution<> dd{param.second, distribution_stddev};
                                float val = dd(gen);
                                MPI_Bcast( &val, 1, MPI_FLOAT, 0, lib4neuro::mpi_active_comm);
                                func->set_parameter(param.first,
                                                    val);
                            }
                        }
                    }
                }
            }

            /* Re-compute coordinates for the new set of ACSFs */
            try {
                this->data_set = this->reader->get_acsf_data_set(*this->net->get_elements()).get();
                this->error_func->set_dataset(this->data_set);
            } catch(const std::runtime_error& err) {
                continue;
            }

            /* Train the network */
            MSE mse(&nn,
                    this->data_set);

            std::vector<double> new_weights = arma::conv_to<std::vector<double>>::from(
                arma::solve(*this->data_set->get_inputs_matrix(),
                            *this->data_set->get_outputs_matrix(),
                            arma::solve_opts::fast));
            /* average over all processes, ASSUMES ALL PROCESSES HAVE SOME DATA */
            MPI_Allreduce( MPI_IN_PLACE, &new_weights[0], new_weights.size(), MPI_DOUBLE, MPI_SUM, lib4neuro::mpi_active_comm );
            for( size_t k = 0; k < new_weights.size(); ++k ){
                new_weights[ k ] /= lib4neuro::mpi_nranks;
            }
            nn.copy_parameter_space(&new_weights);

            current_error = mse.eval();
            MPI_Allreduce( MPI_IN_PLACE, &current_error, 1, MPI_DOUBLE, MPI_SUM, lib4neuro::mpi_active_comm );

            /* Termination condition */
            if(current_error < termination_threshold) {
//                /* Compute correlation of ACSFs */
//                arma::Mat<double> inp_mat = *this->data_set->get_inputs_matrix();
//                arma::Mat<double> m = arma::cor(inp_mat);
////                m.print();
//
//                arma::Col<arma::uword> v = arma::find(m > 0.9);
//                std::unordered_set<std::vector<unsigned long long>, boost::hash<std::vector<unsigned long long>>> correlation_indices;
//                for(size_t k = 0; k < v.size(); k++) {
//                    std::vector<unsigned long long> v_tmp = {v.at(k) % m.n_rows, (unsigned long long)(v.at(k) / m.n_rows)};
//                    if(v_tmp.at(0) == v_tmp.at(1)) {
//                        continue;
//                    }
//                    std::sort(v_tmp.begin(), v_tmp.end());
//                    correlation_indices.insert(v_tmp);
//                }
//
//                if(!correlation_indices.empty()) {
//                    std::cout <<"WARNING: Following " << correlation_indices.size() << " symmetry function pairs correlate significantly (>0.9):" << std::endl;
//                    unsigned int corr_ind = 0;
//                    for(auto e : correlation_indices) {
//                        std::cout << "{" << e.at(0) << ", " << e.at(1) << "} ";
//                        if(corr_ind % 10 == 9) {
//                            std::cout << std::endl;
//                        }
//                        corr_ind++;
//                    }
//                    std::cout << std::endl;
//                } else {
//                    COUT_INFO("Good news! ACSF descriptors do not seem to be correlated.");
//                }
//
//                /* Print range (min & max) of every descriptor */
//                COUT_INFO("Descriptors' ranges:");
//                for(arma::uword k = 0; k < inp_mat.n_cols; k++) {
//                    std::cout << k << ": <" << inp_mat.col(k).min() << ", " << inp_mat.col(k).max() << ">, ";
//                    if(k % 4 == 3) {
//                        std::cout << std::endl;
//                    }
//                }
//
                COUT_INFO("ACSF parameters were optimized after " + std::to_string(iter_ind+1) + " iterations. "
                          "Error of a linear model is: " + std::to_string(current_error) + ".");

                return;
            }

            delta_error   = current_error - previous_error;

            /* Decide about acceptance/rejection of ACSF parameters */
            std::bernoulli_distribution bernoulli_d(std::exp(-delta_error / temperature));
            // arma::Mat<double>* stability_mat = this->data_set->get_stability_matrix();
            // unsigned long stability_coef_max = stability_mat->max();
            // unsigned long long ind_max = stability_mat->index_max();
			unsigned long stability_coef_max = 0;

            if (!(delta_error <= 0 || (delta_error > 0 && (bool)bernoulli_d(gen))) || stability_coef_max > 10) {
//            if (delta_error > 0 || stability_coef_max > 10) {
                // COUT_INFO("Error difference: "
                          // + std::to_string(delta_error) + ", Maximal stability coefficient ("
                          // + std::to_string(ind_max % stability_mat->n_rows) + ","
                          // + std::to_string((unsigned int)(ind_max / stability_mat->n_rows))
                          // + "): " + std::to_string(stability_mat->max()) + ". Reverting parameters' change...")

                /* If conditions are not met, revert the parameter changes */
                for (auto& el : *this->net->get_elements()) {
                    for (auto& func : *el.second->getSymmetryFunctions()) {
                        for (auto& param : *func->get_parameters()) {
                            /* Check, if the parameter is supposed to be optimized */
                            if (std::find(fitted_params.begin(),
                                          fitted_params.end(),
                                          param.first) != fitted_params.end()) {
                                func->revert_parameter_change(param.first);
                            }
                        }
                    }
                }
            } else {
                COUT_INFO("Current error: " + std::to_string(current_error) + " Previous error: " + std::to_string(previous_error))
                previous_error = current_error;
            }

            iter_ind++;
        }

        temperature *= temperature_change;
    }

    COUT_INFO("Parameters optimization reached the maximal number of iterations (" + std::to_string(iter_ind-1)
              + "). The final error of a linear model is " + std::to_string(current_error) + ".");
}
