//
// Created by martin on 20.08.19.
//

#define ARMA_ALLOW_FAKE_GCC

#include <4neuro.h>
#include <iomanip>

void optimize_via_particle_swarm(l4n::NeuralNetwork& net,
                                 l4n::ErrorFunction& ef) {

    /* TRAINING METHOD SETUP */
    std::vector<double> domain_bounds(2 * (net.get_n_weights() + net.get_n_biases()));

    for (size_t i = 0; i < domain_bounds.size() / 2; ++i) {
        domain_bounds[2 * i]     = -150;
        domain_bounds[2 * i + 1] = 150;
    }

    double c1          = 1.7;
    double c2          = 1.7;
    double w           = 0.7;
    size_t n_particles = 300;
    size_t iter_max    = 500;

    /* if the maximal velocity from the previous step is less than 'gamma' times the current maximal velocity, then one
     * terminating criterion is met */
    double gamma = 0.5;

    /* if 'delta' times 'n' particles are in the centroid neighborhood given by the radius 'epsilon', then the second
     * terminating criterion is met ('n' is the total number of particles) */
    double epsilon = 0.02;
    double delta   = 0.7;

    l4n::ParticleSwarm swarm_01(
        &domain_bounds,
        c1,
        c2,
        w,
        gamma,
        epsilon,
        delta,
        n_particles,
        iter_max
    );
    swarm_01.optimize(ef);


    /* ERROR CALCULATION */
    std::cout << "Run finished! Error of the network[Particle swarm]: " << ef.eval(nullptr) << std::endl;
    std::cout
        << "***********************************************************************************************************************"
        << std::endl;
}

double optimize_via_gradient_descent(l4n::NeuralNetwork& net,
                                     l4n::ErrorFunction& ef) {

	if(l4n::mpi_rank == 0){
		std::cout
			<< "***********************************************************************************************************************"
			<< std::endl;
	}
    l4n::GradientDescentBB gd(1e-6,
                              1000,
                              10);

    gd.optimize(ef);


    /* ERROR CALCULATION */
    double err = ef.eval(nullptr);
	if(l4n::mpi_rank == 0){
		std::cout << "Run finished! Error of the network[Gradient descent]: " << err << std::endl;
	}

    /* Just for validation test purposes - NOT necessary for the example to work! */
    return err;
}

double optimize_via_LBMQ(l4n::NeuralNetwork& net,
                         l4n::ErrorFunction& ef,
                         unsigned int niters,
                         unsigned int nb,
                         double tol
                         ) {

    size_t max_iterations = niters;
    size_t batch_size = nb;
    double tolerance = tol;
    double tolerance_gradient = tol * 1e-6;
    double tolerance_parameters = tolerance_gradient;

	if(l4n::mpi_rank == 0){
		std::cout
			<< "***********************************************************************************************************************"
			<< std::endl;
	}
    l4n::LevenbergMarquardt lm(
        max_iterations,
        batch_size,
        tolerance,
        tolerance_gradient,
        tolerance_parameters
    );

     double grate = 1e-2;
     l4n::SmoothingLearning sl( lm, tolerance, grate );

//     sl.optimize(ef);
	lm.optimize(ef);


    /* ERROR CALCULATION */
    double err = ef.eval(nullptr);
    // std::cout << "Run finished! Error of the network[Levenberg-Marquardt]: " << err << std::endl;

    /* Just for validation test purposes - NOT necessary for the example to work! */
    return err;
}

double optimize_via_NormalizedGradientDescent(
	l4n::NeuralNetwork& net,
    l4n::ErrorFunction& ef,
    unsigned int niters,
    unsigned int nb,
    double tol
) {

    size_t max_iterations = niters;
    size_t batch_size = nb;
    double tolerance = tol * 1e-6;

	if(l4n::mpi_rank == 0){
		std::cout
			<< "***********************************************************************************************************************"
			<< std::endl;
	}
    l4n::NormalizedGradientDescent ngd(
        tolerance,
        max_iterations,
        batch_size
    );

    ngd.optimize(ef);


    /* ERROR CALCULATION */
    double err = ef.eval(nullptr);
    // std::cout << "Run finished! Error of the network[Levenberg-Marquardt]: " << err << std::endl;

    /* Just for validation test purposes - NOT necessary for the example to work! */
    return err;
}

double optimize_via_NelderMead(l4n::NeuralNetwork& net, l4n::ErrorFunction& ef) {
    l4n::NelderMead nm(500, 150);

    nm.optimize(ef);

    /* ERROR CALCULATION */
    double err = ef.eval(nullptr);
    std::cout << "Run finished! Error of the network[Nelder-Mead]: " << err << std::endl;

    /* Just for validation test purposes - NOT necessary for the example to work! */
    return err;

}

void dimer_test(
    unsigned int niters,
    unsigned int batch_size,
    double tol,
    std::vector<unsigned int> net_complx,
    const char* src_file,
    const char* src_net_file,
    const char* tgt_net_file,
    const std::vector<double> &g2_cutoff_coefficients,
    const std::vector<double> &g2_extensions,
    const std::vector<double> &g2_shifts,
    const std::vector<double> &g5_cutoff_coefficients,
    const std::vector<double> &g5_extensions,
    const std::vector<bool> &g5_shiftsmax,
    const std::vector<double> &g5_angles
){

    /* Specify cutoff functions */
    std::vector<l4n::CutoffFunction2> g2_cutofffunctions;
    for( auto el: g2_cutoff_coefficients ){
        g2_cutofffunctions.emplace_back( l4n::CutoffFunction2(el) );
    }
    std::vector<l4n::CutoffFunction2> g5_cutofffunctions;
    for( auto el: g5_cutoff_coefficients ){
        g5_cutofffunctions.emplace_back( l4n::CutoffFunction2(el) );
    }

    /* the used symmetry functions */
    std::vector<l4n::SymmetryFunction*> helium_sym_funcs(g2_extensions.size() + g5_extensions.size() );
    /* Specify G2 symmetry functions */
    std::vector<l4n::G2> G2functions;
    for( size_t i = 0; i < g2_extensions.size(); ++i ){
        G2functions.emplace_back( l4n::G2(&g2_cutofffunctions[ i ], g2_extensions[ i ], g2_shifts[ i ] ) );
    }

    /* Specify G2 symmetry functions */
    std::vector<l4n::G5> G5functions;
    auto shift = G2functions.size();
    for( size_t i = 0; i < g5_extensions.size(); ++i ){
        G5functions.emplace_back( l4n::G5(&g5_cutofffunctions[ i ], g5_extensions[ i ], g5_shiftsmax[ i ], g5_angles[ i ] ) );
    }

    for( size_t i = 0; i < g2_extensions.size(); ++i ){
        helium_sym_funcs[ i ] = &G2functions[ i ];
	}
    for( size_t i = 0; i < g5_extensions.size(); ++i ){
        helium_sym_funcs[ i + shift ] = &G5functions[ i ];
    }




    l4n::Element helium = l4n::Element("He",helium_sym_funcs);
    std::unordered_map<l4n::ELEMENT_SYMBOL, l4n::Element*> elements;
    elements[l4n::ELEMENT_SYMBOL::He] = &helium;

    /* Read data */
    l4n::XYZReader reader(src_file, true);
    reader.read();

    if( l4n::mpi_rank == 0 ){
        std::cout << "Finished reading data" << std::endl;
    }

    std::shared_ptr<l4n::DataSet> ds = reader.get_acsf_data_set(elements);
	// ds->print_data();

    /* Create a neural network */
    std::unordered_map<l4n::ELEMENT_SYMBOL, std::vector<unsigned int>> n_hidden_neurons;
    n_hidden_neurons[l4n::ELEMENT_SYMBOL::He] = net_complx;

    std::unordered_map<l4n::ELEMENT_SYMBOL, std::vector<l4n::NEURON_TYPE>> type_hidden_neurons;
    for( auto el :n_hidden_neurons){
        for( auto i = 0; i < el.second.size() - 1; ++i ){
            type_hidden_neurons[el.first].push_back(l4n::NEURON_TYPE::LOGISTIC);
        }
        type_hidden_neurons[el.first].push_back(l4n::NEURON_TYPE::LINEAR);
    }

//    std::unique_ptr<l4n::ACSFNeuralNetwork> net;
//    if( std::strlen(src_net_file) > 0 ){
//        net = std::make_unique<l4n::ACSFNeuralNetwork>(l4n::ACSFNeuralNetwork(src_net_file));
//    }
//    else{
//        net = std::make_unique<l4n::ACSFNeuralNetwork>(l4n::ACSFNeuralNetwork(elements, *reader.get_element_list(), reader.contains_charge(), n_hidden_neurons, type_hidden_neurons));
//    }
//    l4n::MSE mse(&net->get(), ds.get(), false);
    l4n::ACSFNeuralNetwork net(elements, *reader.get_element_list(), reader.contains_charge() && !reader.get_ignore_charge(), n_hidden_neurons, type_hidden_neurons);
    l4n::MSE mse(&net, ds.get(), false);


    net.randomize_parameters();




    double err1 = optimize_via_LBMQ( net, mse, niters, batch_size, tol );
//    double err3 = optimize_via_NormalizedGradientDescent( net, mse, niters, batch_size, tol );
//    double err2 = optimize_via_gradient_descent(net, mse);

    /* Print fit comparison with real data */
    std::vector<double> output;
    output.resize(1);

    double max_error_abs = 0.0, max_error_rel = 0.0;

    std::vector<std::vector<double>> out_dat;
    for(auto e : *mse.get_dataset()->get_data()) {
        net.eval_single(e.first, output);
        double error_= std::abs(e.second.at(0) - output.at(0));
        double error_rel = error_ / std::abs(output.at(0));

        out_dat.push_back({e.first[0], e.first[1], e.second[0], output[0], error_, error_rel});

        if( max_error_abs < error_ ){
            max_error_abs = error_;
        }

        if( error_rel > max_error_rel ){
            max_error_rel = error_rel;
        }
    }

    MPI_Allreduce( MPI_IN_PLACE, &max_error_abs, 1, MPI_DOUBLE, MPI_MAX, l4n::mpi_active_comm );
    MPI_Allreduce( MPI_IN_PLACE, &max_error_rel, 1, MPI_DOUBLE, MPI_MAX, l4n::mpi_active_comm );

    if( l4n::mpi_rank == 0 ){
        std::cout << "*****************************************************" << std::endl;
        std::cout << "maximal absolute error: " << max_error_abs << std::endl;
        std::cout << "maximal relative error: " << 100 * max_error_rel << " %" << std::endl;
        std::cout << "*****************************************************" << std::endl;
        net.save_text(tgt_net_file);
    }

    std::sort(out_dat.begin(), out_dat.end());
    for( int i = 0; i < l4n::mpi_nranks; ++i ){
        if( i == l4n::mpi_rank ){
            for( auto el: out_dat ){
                double error_= el[4];
                double error_rel = el[5];

                if( error_rel > 0.05 ){
                    std::cout << std::setprecision(16) << "{" << el[0] << "} -> {" << el[2] << "} error " << 100*error_rel << " %" << std::endl;
                }
            }
        }
        MPI_Barrier( l4n::mpi_active_comm );
    }

}


int main( int argc, char** argv ) {
    MPI_INIT

    unsigned int niters = 1000;
    unsigned int batch_size = 0;
    double tol = 1e-1;
    char *src_file = "/mem.txt";
    char *src_net_file = "";
    char *tgt_net_file = "/mem.txt";
    unsigned int nlayers = 0;
    std::vector<unsigned int> net_complx = {2, 1};

    size_t id = 1;
    if( argc > id ){
        niters = atoi( argv[id] );
    }
    id++;

    if( argc > id ){
        batch_size = atoi( argv[id] );
    }
    id++;

    if( argc > id ){
        tol = atof( argv[id] );
    }
    id++;

    if( argc > id ){
        src_file = argv[id];
    }
    id++;

    if( argc > id ){
        tgt_net_file = argv[id];
    }
    id++;

    if( argc > id ){
        nlayers = atoi( argv[id] );
        net_complx.resize(nlayers);
    }
    id++;

    for( unsigned int lidx = id; lidx < id + nlayers; ++lidx ){
        net_complx[lidx - id] = atoi( argv[lidx] );
    }
    id += nlayers;

    std::vector<double> g2_cutoff_coefficients = {};
    std::vector<double> g2_extensions = {};
    std::vector<double> g2_shifts = {};
    unsigned int ng2 = 0;
    if( argc > id ){
        ng2 = atoi( argv[id] );
        g2_extensions.resize(ng2);
        g2_shifts.resize(ng2);
        g2_cutoff_coefficients.resize(ng2);
    }
    id++;

    for( unsigned int sidx = id; sidx < id + ng2; ++sidx ){
        g2_cutoff_coefficients[sidx - id] = atof( argv[sidx] );
    }
    id += ng2;
    for( unsigned int sidx = id; sidx < id + ng2; ++sidx ){
        g2_extensions[sidx - id] = atof( argv[sidx] );
    }
    id += ng2;
    for( unsigned int sidx = id; sidx < id + ng2; ++sidx ){
        g2_shifts[sidx - id] = atof( argv[sidx] );
    }
    id += ng2;

    std::vector<double> g5_cutoff_coefficients = {};
    std::vector<double> g5_extensions = {};
    std::vector<bool> g5_shifts = {};
    std::vector<double> g5_angles = {};
    unsigned int ng5 = 0;
    if( argc > id ){
        ng5 = atoi( argv[id] );
        g5_extensions.resize(ng5);
        g5_shifts.resize(ng5);
        g5_angles.resize(ng5);
        g5_cutoff_coefficients.resize(ng5);
    }
    id++;

    for( unsigned int sidx = id; sidx < id + ng5; ++sidx ){
        g5_cutoff_coefficients[sidx - id] = atof( argv[sidx] );
    }
    id += ng5;
    for( unsigned int sidx = id; sidx < id + ng5; ++sidx ){
        g5_extensions[sidx - id] = atof( argv[sidx] );
    }
    id += ng5;
    for( unsigned int sidx = id; sidx < id + ng5; ++sidx ){
        g5_shifts[sidx - id] = (bool)atoi( argv[sidx] );
    }
    id += ng5;
    for( unsigned int sidx = id; sidx < id + ng5; ++sidx ){
        g5_angles[sidx - id] = atof( argv[sidx] );
    }
    id += ng5;

    if( argc > id ){
        src_net_file = argv[id];
    }
    id++;


    nlayers = net_complx.size();
    ng2 = g2_extensions.size();
    ng5 = g5_extensions.size();

    if( l4n::mpi_rank == 0 ){

        std::cout << "Maximal number of iterations: " << niters << std::endl;
        std::cout << "Batch size: " << batch_size << std::endl;
        std::cout << "Learning tolerance: " << tol << std::endl;

        std::cout << "# of network layers: " << nlayers << std::endl;
        std::cout << "# of neurons in the layers: ";
        for( auto el: net_complx ){
            std::cout << el << " ";
        }
        std::cout << std::endl;
        if( ng2 > 0 ){
            std::cout << "***********************************************************" << std::endl;
            std::cout << "number of G2 symmetry functions: " << ng2 << std::endl;
            for( size_t i = 0; i < ng2; ++i ){
                std::cout << " f" << i << ": cutoff " << g2_cutoff_coefficients[ i ] << ", extension " << g2_extensions[ i ] << ", shift " << g2_shifts[ i ] << std::endl;
            }
        }
        if( ng5 > 0 ){
            std::cout << "***********************************************************" << std::endl;
            std::cout << "number of G5 symmetry functions: " << ng5 << std::endl;
            for( size_t i = 0; i < ng5; ++i ){
                std::cout << " f" << i << ": cutoff " << g5_cutoff_coefficients[ i ] << ", extension " << g5_extensions[ i ] << ", shift " << g5_shifts[ i ] << ", angle " << g5_angles[ i ] << std::endl;
            }
        }
        std::cout << "***********************************************************" << std::endl;
        std::cout << "Data file: " << src_file << std::endl;
        std::cout << "Network will be loaded from: " << src_net_file << std::endl;
        std::cout << "Network will be saved in: " << tgt_net_file << std::endl;
        std::cout << "***********************************************************" << std::endl << std::endl;
    }
    try{

        dimer_test(
            niters,
            batch_size,
            tol,
            net_complx,
            src_file,
            src_net_file,
            tgt_net_file,
            g2_cutoff_coefficients,
            g2_extensions,
            g2_shifts,
            g2_cutoff_coefficients,
            g5_extensions,
            g5_shifts,
            g5_angles
        );

    } catch (const std::exception& e) {
        std::cerr << e.what() << std::endl;
        exit(EXIT_FAILURE);
    }

    MPI_FINISH
    return 0;
}
