/**
 * DESCRIPTION OF THE CLASS
 *
 * @author David Vojtek
 * @date 2018
 */

#define BOOST_TEST_MODULE NeuralNetwork_test

// TODO fix boost_test_lib and remove the following include!
#ifdef _WINDOWS
#include <boost/test/included/unit_test.hpp>
#endif

#include "boost_unit_tests_preamble.h"
#include <turtle/mock.hpp>
#include "../Network/NeuralNetwork.h"

using namespace lib4neuro;


MOCK_BASE_CLASS(mock_NeuronLinear,
                lib4neuro::NeuronLinear
) {
    MOCK_METHOD(activate,
                2)

    MOCK_METHOD(activation_function_eval_derivative_bias,
                2)

    MOCK_METHOD(activation_function_eval_derivative,
                2)

    MOCK_METHOD(get_derivative,
                0)
};

/**
 * Boost testing suite for testing NeuralNetwork.h
 */
BOOST_AUTO_TEST_SUITE(NeuralNetwork_test)

/**
 * Test of creating new instance of NeuralNetwork
 */
    BOOST_AUTO_TEST_CASE(NeuralNetwork_construction_test) {
        //Test of none exception raise when creating new instance of NeuralNetwork
        BOOST_CHECK_NO_THROW(NeuralNetwork network);
    }

/**
 * Test of add_neuron method
 * Existing bias out of range cancelation
 */
    BOOST_AUTO_TEST_CASE(NeuralNetwork_add_neuron_test) {
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear);
        std::shared_ptr<mock_NeuronLinear> n2(new mock_NeuronLinear);
        std::shared_ptr<mock_NeuronLinear> n3(new mock_NeuronLinear);
        std::shared_ptr<mock_NeuronLinear> n4(new mock_NeuronLinear);

        NeuralNetwork network;

        //Tests of correct neuron indices when add_neuron
        BOOST_CHECK_EQUAL(0,
                          network.add_neuron(n1));

        BOOST_CHECK_EQUAL(1,
                          network.add_neuron(n2,
                                             BIAS_TYPE::NEXT_BIAS));

        BOOST_CHECK_EQUAL(2,
                          network.add_neuron(n3,
                                             BIAS_TYPE::NO_BIAS));

        BOOST_CHECK_EQUAL(3,
                          network.get_n_neurons());

        BOOST_CHECK_EQUAL(2,
                          network.get_n_biases());

    }

/**
 * Test of add_connection_simple method
 */
    BOOST_AUTO_TEST_CASE(NeuralNetwork_add_connection_simple_test) {
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear);
        std::shared_ptr<mock_NeuronLinear> n2(new mock_NeuronLinear);
        NeuralNetwork                      network;
        network.add_neuron(n1,
                           BIAS_TYPE::NO_BIAS);
        network.add_neuron(n2,
                           BIAS_TYPE::NO_BIAS);

        BOOST_CHECK_EQUAL(0,
                          network.add_connection_simple(0,
                                                        1,
                                                        SIMPLE_CONNECTION_TYPE::NEXT_WEIGHT));
        BOOST_CHECK_EQUAL(1,
                          network.add_connection_simple(0,
                                                        1,
                                                        SIMPLE_CONNECTION_TYPE::NEXT_WEIGHT,
                                                        5));
        BOOST_CHECK_EQUAL(2,
                          network.add_connection_simple(0,
                                                        1,
                                                        SIMPLE_CONNECTION_TYPE::UNITARY_WEIGHT));
        BOOST_CHECK_EQUAL(3,
                          network.add_connection_simple(0,
                                                        1,
                                                        SIMPLE_CONNECTION_TYPE::UNITARY_WEIGHT,
                                                        5));
        BOOST_CHECK_EQUAL(4,
                          network.add_connection_simple(0,
                                                        1,
                                                        SIMPLE_CONNECTION_TYPE::EXISTING_WEIGHT,
                                                        1));

        BOOST_CHECK_EQUAL(2,
                          network.get_n_weights());
    }

/**
 * Test of add_connection_general method
 */
    BOOST_AUTO_TEST_CASE(NeuralNetwork_specify_inputs_neurons_test) {
        NeuralNetwork                      network;
        mock_NeuronLinear                  po;
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear());

        network.add_neuron(n1,
                           BIAS_TYPE::NO_BIAS);

        std::vector<size_t> inputs;
        inputs.push_back(0);

        BOOST_CHECK_EQUAL(0,
                          network.get_n_inputs());
        network.specify_input_neurons(inputs);
        BOOST_CHECK_EQUAL(1,
                          network.get_n_inputs());
    }

    BOOST_AUTO_TEST_CASE(NeuralNetwork_specify_outputs_neurons_test) {
        NeuralNetwork                      network;
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear);
        network.add_neuron(n1,
                           BIAS_TYPE::NO_BIAS);

        std::vector<size_t> outputs;
        outputs.push_back(0);

        BOOST_CHECK_EQUAL(0,
                          network.get_n_outputs());
        network.specify_output_neurons(outputs);
        BOOST_CHECK_EQUAL(1,
                          network.get_n_outputs());
    }

    BOOST_AUTO_TEST_CASE(NeuralNetwork_eval_single_test) {
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear());
        std::shared_ptr<mock_NeuronLinear> n2(new mock_NeuronLinear());

        mock_NeuronLinear n3 = *n1;
        mock_NeuronLinear n4 = *n2;
        MOCK_EXPECT(n3.activate).returns(5);
        MOCK_EXPECT(n4.activate).returns(5);


        NeuralNetwork network;
        network.add_neuron(n1);
        network.add_neuron(n2);

        network.add_connection_simple(0,
                                      1,
                                      SIMPLE_CONNECTION_TYPE::UNITARY_WEIGHT,
                                      2.5);

        std::vector<size_t> output_neuron_indices;
        output_neuron_indices.push_back(1);
        network.specify_output_neurons(output_neuron_indices);

        std::vector<size_t> input_neuron_indices;
        input_neuron_indices.push_back(0);
        network.specify_input_neurons(input_neuron_indices);

        std::vector<double> input;
        input.push_back(5);
        std::vector<double> output;
        output.push_back(8);

        network.eval_single(input,
                            output);
        BOOST_CHECK_EQUAL(5,
                          output.at(0));
    }

    BOOST_AUTO_TEST_CASE(NeuralNetwork_randomize_weights_test) {
        std::shared_ptr<mock_NeuronLinear> n1(new mock_NeuronLinear());
        std::shared_ptr<mock_NeuronLinear> n2(new mock_NeuronLinear());

        NeuralNetwork network;
        network.add_neuron(n1,
                           BIAS_TYPE::NO_BIAS);
        network.add_neuron(n2,
                           BIAS_TYPE::NO_BIAS);

        for (int i = 0; i < 100; i++) {
            network.add_connection_simple(0,
                                          1,
                                          SIMPLE_CONNECTION_TYPE::NEXT_WEIGHT);
        }
        network.randomize_weights();
        std::vector<double>* weights = network.get_parameter_ptr_weights();

        double sum = 0;

        for (int i = 0; i < 100; i++) {
            sum += weights->at(i);
        }
        sum        = sum / 100;
        BOOST_CHECK(sum < 0.15 && sum > -0.15);
    }

BOOST_AUTO_TEST_SUITE_END()
